
/* simple utility to convert an .mdk file output by autobusk to a standard
   .mid file; written 2000-05-10, overhauled 2014-12-11 by Albert Graef */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

typedef unsigned long ulong;
typedef unsigned char ubyte;

/* anything which delimits path from filename */

#define DIRSTR "/\\:"

/* return filename with new extension */

char *getext(char *s)
{
  char *t = strrchr(s, '.'), *p;
  if (t) {
    for (p = t; *p; p++)
      if (strchr(DIRSTR, *p))
	return "";
    return t;
  } else
    return "";
}

char *newext(char *s, char *next)
{
  char *ext;
  if ((ext = getext(s))) {
    int l = strlen(s), e = strlen(ext);
    char *t = malloc(l-e+strlen(next)+1);
    if (!t) return NULL;
    strncpy(t, s, l-e);
    strcpy(t+l-e, next);
    return t;
  } else {
    char *t = malloc(strlen(s)+strlen(next)+1);
    strcat(strcpy(t, s), next);
    return t;
  }
}

/* file routines ************************************************************/

/* following routines return nonzero in case of an error */

/* parse a line of an mdk file (only P and D commands are recognized,
   rest and rubbish is silently ignored); return -1 iff at end of file */

ulong curtime = 0;

int parse_cmd(FILE *fp, ulong *delta, ubyte *cmd, ubyte *byte1, ubyte *byte2)
{
 next:
  if (feof(fp))
    return -1;
  else if (ferror(fp))
    return -2;
  else {
    ulong newtime, secs, msecs, val1, val2;
    char track, cmdc, buf[1000];
    int args;

    if (!fgets(buf, sizeof(buf), fp)) {
      if (feof(fp))
	return -1;
      else
	return -2;
    }
    if (buf[strlen(buf)-1] != '\n')
      return -2; /* buffer overflow */
    if ((args = sscanf(buf, "%lu %lu %c %c %lu %lu", &secs, &msecs,
		       &track, &cmdc, &val1, &val2)) < 5
	|| track < 'a' || track > 'p'
	|| (cmdc != 'P' && cmdc != 'D')
	|| (cmdc == 'P' && args < 6)
	|| val1 > 0xFF || val2 > 0xFF)
      goto next; /* ignore */
    newtime = secs*1000+msecs;
    if (newtime < curtime) goto next;
    *delta = newtime - curtime;
    *cmd = ((cmdc == 'D')?0x80:0x90)+(track-'a');
    *byte1 = val1;
    if (args == 6)
      *byte2 = val2;
    else
      *byte2 = 0;
    curtime = newtime;
    return 0;
  }
}

/* write a variable length delta value to a midi stream (pilfered from MIDI
   Technical Fanatic's Brainwashing Center) */

ulong nbytes = 0;

void write_var_len(register FILE *outfile, register ulong value)
{
  register unsigned long buffer;
  buffer = value & 0x7F;
  
  while ( (value >>= 7) )
    {
      buffer <<= 8;
      buffer |= ((value & 0x7F) | 0x80);
    }

  while (1)
    {
      putc(buffer,outfile); nbytes++;
      if (buffer & 0x80)
	buffer >>= 8;
      else
	break;
    }
}

/* spit out a single (note on or off) MIDI command */

int write_cmd(FILE *fp, ulong delta, ubyte cmd, ubyte byte1, ubyte byte2)
{
/*   printf("%4lu: $%2X %3u %3u\n", delta, cmd, byte1, byte2); */
  write_var_len(fp, delta);
  putc(cmd, fp); nbytes++;
  putc(byte1, fp); nbytes++;
  putc(byte2, fp); nbytes++;
  if (ferror(fp))
    return -1;
  else
    return 0;
}

/* write a header for a type 1 midi file with millisecond deltas */

/* NOTE: As most MIDI players, DAWs and sequencers don't know how to deal with
   milliseconds divisions (E7 28), we use a division of 500 ppqn instead,
   which equals milliseconds time stamps at a tempo of 120 bpm. This should
   work out of the box with most MIDI players, but to make sure that even the
   stupidest players get it right, we also add an appropriate tempo track. */

ubyte hdr[] = {
  0x4D, 0x54, 0x68, 0x64,     /* MThd ID */
  0x00, 0x00, 0x00, 0x06,     /* Length of the MThd chunk is always 6. */
  0x00, 0x01,                 /* The Format type is 1. */
  0x00, 0x02,                 /* There are 2 MTrk chunks in this file. */
  0x01, 0xF4,                 /* Division: 500 ppqn. */

  0x4D, 0x54, 0x72, 0x6B,     /* MTrk ID */
  0x00, 0x00, 0x00, 0x0B,     /* Length 11. */
  0x00,                       /* Delta */
  0xFF, 0x51, 0x03,           /* Tempo */
  0x07, 0xA1, 0x20,           /* 120 bpm (500000 µs per quarter note). */
  0x00, 0xFF, 0x2F, 0x00,     /* End of track. */

  0x4D, 0x54, 0x72, 0x6B,     /* MTrk ID */
  0x00, 0x00, 0x00, 0x00,     /* Length (fixed when file has been written). */
};

int write_header(FILE *fp)
{
  if (write(fileno(fp), hdr, sizeof(hdr)) < sizeof(hdr))
    return -1;
  else
    return 0;
}

/* fix length information in header when whole track has been written */

int fix_header(FILE *fp)
{
  int len;
  rewind(fp);
  /* fill in length (number of MTrk bytes written) in bigendian order */
  len = sizeof(hdr);
  hdr[--len] = nbytes & 0xFF; nbytes >>= 8;
  hdr[--len] = nbytes & 0xFF; nbytes >>= 8;
  hdr[--len] = nbytes & 0xFF; nbytes >>= 8;
  hdr[--len] = nbytes & 0xFF; nbytes >>= 8;
  /* rewrite the header */
  return write_header(fp);
}

/* write end of track marker */

int write_eot(FILE *fp)
{
  putc(0x00, fp); nbytes++;
  putc(0xFF, fp); nbytes++;
  putc(0x2F, fp); nbytes++;
  putc(0x00, fp); nbytes++;
  if (ferror(fp))
    return -1;
  else
    return 0;
}

/* main program *************************************************************/

int main(int argc, char *argv[])
{
  FILE *infp, *outfp;
  char *midfile;
  ulong delta;
  ubyte cmd, byte1, byte2;
  int res;

  if (argc != 2) {
    fprintf(stderr, "Usage: %s MDK-FILE\n", argv[0]);
    exit(1);
  } else if (!(infp = fopen(argv[1], "r"))) {
    fprintf(stderr, "%s: File %s not found\n", argv[0], argv[1]);
    exit(1);
  }
  midfile = newext(argv[1], ".mid");
  if (!midfile) {
    fprintf(stderr, "%s: Memory overflow\n", argv[0]);
    fclose(infp);
  }
  /* on MSDOS/Windows, output file must be opened in binary mode! */
  if (!(outfp = fopen(midfile, "wb"))) {
    fprintf(stderr, "%s: Error creating %s\n", argv[0], midfile);
    fclose(infp);
    exit(1);
  }
  if (write_header(outfp)) {
    fprintf(stderr, "%s: Error writing %s\n", argv[0], midfile);
    goto errexit;
  }
  while (!(res = parse_cmd(infp, &delta, &cmd, &byte1, &byte2)))
    if (write_cmd(outfp, delta, cmd, byte1, byte2)) {
      fprintf(stderr, "%s: Error writing %s\n", argv[0], midfile);
      goto errexit;
    }
  if (res != -1) {
    fprintf(stderr, "%s: Error reading %s\n", argv[0], argv[1]);
    goto errexit;
  }
  if (write_eot(outfp) || fix_header(outfp)) {
    fprintf(stderr, "%s: Error writing %s\n", argv[0], midfile);
    goto errexit;
  }
  fclose(infp); fclose(outfp);
  exit(0);
 errexit:
  fclose(infp); fclose(outfp);
  unlink(midfile);
  exit(1);
}
